
# lineFilterer.py
# Filters two-file corpora like those available from EUROPARL,
# removing lines that violate certain parameters (eg. too
# many / too few words), re-formatting the acceptable lines
# (eg. removing punctuation, changing to lower-case), and writing
# the filtered results to two new files, based on the command-line
# arguments.

import sys
import re
from optparse import OptionParser

def getLineSubset(firstFilename, firstFilter, secondFilename=None, secondFilter=None, verbose=False, outPrefix=None, stripOutput=False):
    '''Returns the number of lines in the files with filenames
    firstFilename and secondFilename that satisfy the given filters.
    Each line number is satisfactory if firstFilter returns
    true with that line from firstFile as an argument, and if secondFilter
    does the same (if secondFilter and secondFile aren't None).
    If outPrefix is not None, the satisfactory line pairs are written out
    to a file, with the firstFilename lines going to outPrefix+".lang1"
    and the secondFilename lines to outPrefix+".lang2".'''

    if (secondFilename == None) != (secondFilter == None):
        raise RuntimeError("only one of secondFilename and secondFilter specified")

    hasSecond = (secondFilename != None)
    hasOutput = (outPrefix != None)

    firstFile = open(firstFilename, 'r')
    if hasSecond:
        secondFile = open(secondFilename, 'r')

    if hasOutput:
        firstOutFile = open(outPrefix + ".lang1", 'w')
        if hasSecond:
            secondOutFile = open(outPrefix + ".lang2", 'w')

    outList = []
    count = 0
    lineNumber = 1

    nextFirst = firstFile.readline()
    if hasSecond:
        nextSecond = secondFile.readline()
    else:
        nextSecond = "foo"
        secondFilter = (lambda x : True)

    while len(nextFirst) != 0 and len(nextSecond) != 0:
        # examine it for length
        if firstFilter(nextFirst) and secondFilter(nextSecond):
            if stripOutput:
                nextFirst = stripLine(nextFirst)
                if hasSecond:
                    nextSecond = stripLine(nextSecond)
            if verbose:
                print nextFirst,
                if hasSecond:
                    print nextSecond
            if hasOutput:
                firstOutFile.write(nextFirst)
                if hasSecond:
                    secondOutFile.write(nextSecond)
            #outList.append(lineNumber)  # UBERHACK!!!!!!1
            count += 1
        nextFirst = firstFile.readline()
        if hasSecond:
            nextSecond = secondFile.readline()
        lineNumber += 1

    #raw_input("Waiting for input so RAM use can be seen...")

    firstFile.close()
    if hasSecond:
        secondFile.close()

    if hasOutput:
        firstOutFile.close()
        if hasSecond:
            secondOutFile.close()

    return count

# Output stripper

def stripLine(line):
    '''Simplifies the output by casting to lower case
    and removing punctuation.'''

    # this is an ugly approach to filtering out punctuation, but
    # since the corpus may have proper names with accented letters
    # that "string".isalpha() returns False for, it seemed better.
    badChars = r"\"!@#$%^&*()1234567890~`-=_+[]{}\|;:',.<>/ ?	"

    uLine = unicode(line, "utf-8")

    outChars = []
    isWhite = True
    for char in uLine:
        #if char not in badChars:
        if char.isalpha():
            isWhite = False
            outChars.append(char.lower().encode("utf-8"))
        else:
            if not isWhite:
                isWhite = True
                outChars.append(' ')

    if not isWhite:
        return ''.join(outChars) + "\n"
    else:
        return ''.join(outChars[:-1]) + "\n"

# Filters

def worker_maxNumWords(line, maxWords):
    return (len(line.split()) <= maxWords)

def filter_maxNumWords(maxWords, otherFilter=(lambda s: True)):
    return lambda s: (worker_maxNumWords(s, maxWords) and otherFilter(s))

def worker_minNumWords(line, minWords):
    return (len(line.split()) >= minWords)

def filter_minNumWords(minWords, otherFilter=(lambda s: True)):
    return lambda s: (worker_minNumWords(s, minWords) and otherFilter(s))

def worker_removeMatches(line, regexObjectList):
    '''Returns True iff each of the regex objects in regexObjectList
    can find a match in line using search()'''
    for regex in regexObjectList:
        if regex.search(line):
            return False
    return True

def filter_procedural(otherFilter=(lambda s: True)):
    '''Removes procedural notes from the English corpus'''
    regexStringList = [r"Question No .* by (.*)",\
                       r"(.*pplau.*)",\
                       r"VOTE",\
                       r"eport (.*-.*)",\
                       r"^\s*\(.*\)\s*$",\
                       r"^Subject:",\
                       r"^EXPLANATION",\
                       r"^TOPICAL AND",\
                       r"^WRITTEN",\
                       r"^.uestion .ime"]
    #regexStringList = [r"[ ABCDEFGHIJKLMNOPQRSTUVWXYZ]{10,18}"]
    regexObjectList = [re.compile(regex) for regex in regexStringList]

    return lambda s: (worker_removeMatches(s, regexObjectList)\
        and otherFilter(s))

def main():
    usage = "usage: %prog [options] english other"
    parser = OptionParser()
    parser.add_option("-o", "--out-prefix",\
        dest="out_prefix", default=None, metavar="PREFIX",\
        help='output the filtered files to PREFIX.lang1, PREFIX.lang2')
    parser.add_option("-s", "--strip",\
        action="store_true", dest="strip", default=False,\
        help='cast string to lower-case and remove punctuation for output')
    parser.add_option("-m", "--minimum",\
        action="store", type="int", dest="minimum",\
        help="require lines (both langs) to have at least MINIMUM words")
    parser.add_option("-M", "--maximum",\
        action="store", type="int", dest="maximum",\
        help="require lines (both langs) to have at most MAXIMUM words")
    parser.add_option("-p", "--remove-procedural",\
        action="store_true", dest="remove_procedural", default=False,\
        help='remove non-sentence procedural notes from english (eg. "Question No ## by (H-####/##):")')
    parser.add_option("-c", "--remove-common",\
        action="store_true", dest="remove_common", default=False,\
        help='remove manually-identified common sentences from english (eg. "The debate is closed")')
    parser.add_option("-v", "--verbose",\
        action="store_true", dest="verbose", default=False,\
        help='print all satisfactory sentences as they are found')

    (options, args) = parser.parse_args()

    if len(args) == 0:
        parser.print_help()
        exit(1)

    # deal with args


    hasSecond = False
    firstFilename = args[0]
    if len(args) > 1:
        secondFilename = args[1]
        hasSecond = True

    firstFilter = lambda s: True
    secondFilter = lambda s: True

    if options.minimum != None:
        firstFilter = filter_minNumWords(int(options.minimum), firstFilter)
        if hasSecond:
            secondFilter = filter_minNumWords(int(options.minimum), secondFilter)
    if options.maximum != None:
        firstFilter = filter_maxNumWords(int(options.maximum), firstFilter)
        if hasSecond:
            secondFilter = filter_maxNumWords(int(options.maximum), secondFilter)

    if options.remove_procedural:
        firstFilter = filter_procedural(firstFilter)

    if hasSecond:
        lineNumbers = getLineSubset(firstFilename, firstFilter, secondFilename, secondFilter, verbose=options.verbose, outPrefix=options.out_prefix, stripOutput=options.strip)
    else:
        lineNumbers = getLineSubset(firstFilename, firstFilter, verbose=options.verbose, outPrefix=options.out_prefix, stripOutput=options.strip)
    print "Number of good lines:", lineNumbers

if __name__ == "__main__":
    main()
